package net.sf.corn.converter;

import java.util.HashMap;
import java.util.HashSet;

/**
 * This class keeps object references within it. It's designed as a helper class 
 * to detect circular object references for java serialization/transformation purposes
 * @author Serhat Dirik
 */
public class ObjectReferenceMap {

	 private HashMap<Class<?>, HashSet<Integer>> objectMap = new HashMap<Class<?>, HashSet<Integer>>();
	 
	 /**
	  * Puts an object reference to the map.It never thrown an exception for the null references.It simply skips them.
	  * This method uses object.hashCode() method for identity detection
	  * @param obj
	  * @throws ObjectExistException thrown if object already exist in the map
	  */
	 public synchronized void putObject(Object obj) throws ConversionException{
		
		 putObject(obj,findObjectIdentity(obj));
	 }
	 
	 /**
	  * Removes an object reference from the map.It never thrown an exception for the null references.It simply skips them.
	  * This method uses object.hashCode() method for identity detection
	  * @param obj
	  */
	 public synchronized void removeObject(Object obj){
		
		 removeObject(obj,findObjectIdentity(obj));
	 }
	 
	 /**
	  * Puts an object reference to the map.It never thrown an exception for the null references.It simply skips them.
	  * This method uses given referenceId for object identification.Do not use this method unless you really indented to use 
	  * propriety object identities rather than hashCodes of those objects
	  * @param obj
	  * @param refId Object id
	  * @throws ObjectExistException thrown if object already exist in the map
	  */
	 private synchronized void putObject(Object obj,Integer refId) throws ConversionException{
		 if(obj == null) return;
		 HashSet<Integer> identitySet = objectMap.get(obj.getClass());
		 if(identitySet != null){
			 if(identitySet.contains(refId)){
				 throw new ConversionException("Class: "+obj.getClass().getName()+" HashCode:"+findObjectIdentity(obj)+" already exists in the map");
			 }
		 }else{
			 identitySet = new HashSet<Integer>();
		 }
		 identitySet.add(refId);
		 objectMap.put(obj.getClass(),identitySet);
	 }
	 
	 private synchronized void removeObject(Object obj,Integer refId){
		 if(obj == null) return;
		 HashSet<Integer> identitySet = objectMap.get(obj.getClass());
		 if(identitySet != null){
			 if(identitySet.contains(refId)){
				 identitySet.remove(refId);
			 }
		 }else{
			 return;
		 }
		 objectMap.put(obj.getClass(),identitySet);
	 }
	 
	 /**
	  * Checks if the object reference already exist in the map
	  * @param obj
	  * @return true if exist
	  */
	 public boolean contains(Object obj){
		 if(obj == null) return false;
		 HashSet<Integer> identitySet = objectMap.get(obj.getClass());
		 if(identitySet == null) return false;
		 return identitySet.contains(findObjectIdentity(obj));
	 }
	 
	 /**
	  * Checks if the object reference already exist in the map with the given referenceId
	  * @param obj
	  * @return true if exist
	  */
	 public boolean contains(Object obj,int refId){
		 if(obj == null) return false;
		 HashSet<Integer> identitySet = objectMap.get(obj.getClass());
		 if(identitySet == null) return false;
		 return identitySet.contains(refId);
	 }
	 
	 /**
	  * Clears the reference map
	  */
	 public void clear(){
		 objectMap.clear();
	 }
	 
	 private static Integer findObjectIdentity(Object o){
		 if(o == null) return 0;
		 return new Integer(System.identityHashCode (o));
	 }
}
