package net.sf.corn.converter.json;

import java.io.Serializable;
import java.lang.reflect.Array;

/**
 * This is the abstract base class for all json types
 *  * @author Serhat Dirik
 *
 */
public abstract class JsTypeObject implements Serializable {

	private static final long serialVersionUID = 2134562554702042153L;
	
	protected static final transient char CHAR_DOUBLEQUOTE = '"';
	protected static final transient char CHAR_BACKSLASH = '\\';
	protected static final transient char CHAR_BACKSPACE = '\b';
	protected static final transient char CHAR_SLASH = '/';
	protected static final transient char CHAR_SMALLERTHAN = '<';
	protected static final transient char CHAR_NEWLINE = '\n';
	protected static final transient char CHAR_FORMFEED = '\f';
	protected static final transient char CHAR_CARRIAGE_RETURN = '\r';
	protected static final transient char CHAR_TAB = '\t';
	protected static final transient char CHAR_WHITESPACE = ' ';
	protected static final transient char CHAR_CONTROL = '\u0080';
	protected static final transient char CHAR_EN_QUAD = '\u2000';
	protected static final transient char CHAR_ACCOUNT_OF = '\u2100';
	protected static final transient char CHAR_NO_BREAKSPACE = '\u00a0';
	protected static final transient String STR_TRIPLEO = "000";
	protected static final transient String STR_UNICODE_START = "\\u";
	protected static final transient String ESCAPED_TAB = "\\t";
	protected static final transient String ESCAPED_CARRIAGE_RETURN = "\\r";
	protected static final transient String ESCAPED_FORMFEED = "\\f";
	protected static final transient String ESCAPED_NEWLINE = "\\n";
	protected static final transient String ESCAPED_BACKSPACE = "\\b";
	protected static final transient String STR_EMPTY = "\"\"";
	
	public    static final transient String NULL_STRING_VALUE = "null";
	/**
	 * Name of the java class
	 */
	protected String javaClassName = null;

	/**
	 * Gives java class name that we converted from 
	 * @return
	 */
	public String getJavaClassName() {
		return javaClassName;
	}

	/**
	 * Instantiates and returns the java class 
	 * @return the java class that we converted from
	 * @throws ClassNotFoundException throwns if java class not found in the classpath
	 */
	public Class<?> getJavaClass() throws ClassNotFoundException {
		if (getJavaClassName() == null)
			return null;
		String class_name = getJavaClassName();
		boolean isArray = false;
		if (class_name.endsWith("[]")) {
			isArray = true;
			class_name = class_name.substring(0, class_name.indexOf('['));
		}
		Class<?> clazz = Class.forName(class_name);
		if (isArray && size() > 0) {
			Object o = Array.newInstance(clazz, size());
			clazz = o.getClass();
		}
		return clazz;
	}
	/**
	 * Set the java class name that we converted from
	 * @param javaClassName name of the java class
	 */
	public void setJavaClassName(String javaClassName) {
		this.javaClassName = javaClassName;
	}

	/**
	 * Gives the JSON representation of this object
	 * @return JSON representation
	 */
	public abstract String toJsonString();
	
	/**
	 * String value of the contained object 
	 * @return naked string value
	 */
	public abstract String toStringValue();

	/**
	 * Tells whether this object represents JSON Null or not
	 * @return true if represents null
	 */
	public abstract boolean isNull();

	/**
	 * Tells if there is no value set yet
	 * @return true if this object has no value yet
	 */
	public final boolean isEmpty() {
		return isNull() || size() <= 0;
	}
	/**
	 * Tells if this object is complex type or not
	 * @return true if this object is complex type
	 */
	public abstract boolean isComplex();
	/**
	 * Tells if this object is List typed object or not
	 * @return true if this object is a list typed
	 */
	public abstract boolean isList();
	/**
	 * Tells if this object is Simple type object or not. Null objects are also considered as simple
	 * @return true if this object is simple
	 */
	public final boolean isSimple() {
		return !isComplex() && !isList();
	}
	/**
	 * Tells if the object is converted from an Java array type.
	 * @return true if the object is an array typed
	 */
	public final boolean isArray() {
		if (isNull() || !isList())
			return false;
		String class_name = getJavaClassName();
		return (class_name != null && class_name.endsWith("[]"));
	}
	/**
	 * Gives size of the contained object values within the object
	 * @return size of values
	 */
	public abstract int size();

	/**
	 * Produce a string in double quotes with backslash sequences in all the
	 * right places. A backslash will be inserted within </, allowing JSON text
	 * to be delivered in HTML. In JSON text, a string cannot contain a control
	 * character or an unescaped quote or backslash.
	 */
	public static String quote(String string) {
		if (string == null || string.length() == 0) {
			return STR_EMPTY;
		}

		char b;
		char c = 0;
		int i;
		String t;
		int len = string.length();
		StringBuilder sb = new StringBuilder();
		sb.append(CHAR_DOUBLEQUOTE);
		for (i = 0; i < len; i += 1) {
			b = c;
			c = string.charAt(i);
			switch (c) {
			case CHAR_BACKSLASH: // case CHAR_SINGLEQUOTE:
			case CHAR_DOUBLEQUOTE:
				sb.append(CHAR_BACKSLASH);
				sb.append(c);
				break;
			case CHAR_SLASH:
				if (b == CHAR_SMALLERTHAN) {
					sb.append(CHAR_BACKSLASH);
				}
				sb.append(c);
				break;
			case CHAR_BACKSPACE:
				sb.append(ESCAPED_BACKSPACE);
				break;
			case CHAR_TAB:
				sb.append(ESCAPED_TAB);
				break;
			case CHAR_NEWLINE:
				sb.append(ESCAPED_NEWLINE);
				break;
			case CHAR_FORMFEED:
				sb.append(ESCAPED_FORMFEED);
				break;
			case CHAR_CARRIAGE_RETURN:
				sb.append(ESCAPED_CARRIAGE_RETURN);
				break;
			default:
				if (c < CHAR_WHITESPACE
						|| (c >= CHAR_CONTROL && c < CHAR_NO_BREAKSPACE)
						|| (c >= CHAR_EN_QUAD && c < CHAR_ACCOUNT_OF)) {
					t = STR_TRIPLEO + Integer.toHexString(c);
					sb.append(STR_UNICODE_START).append(
							t.substring(t.length() - 4));
				} else {
					sb.append(c);
				}
			}
		}
		sb.append(CHAR_DOUBLEQUOTE);
		return sb.toString();

	}

	/**
	 * Removes the quotes arround the text
	 * @param string quoted text
	 * @return unquoted text
	 */
	public static String unquote(String string) {
		if (string == null || string.trim().length() == 0)
			return string;
		int len = string.length();
		char c = 0;
		StringBuilder sb = new StringBuilder();
		boolean prevback = false;
		for (int i = 0; i < len; i += 1) {
			c = string.charAt(i);
			switch (c) {
			case CHAR_BACKSLASH: // case CHAR_SINGLEQUOTE:
				if(prevback){
					sb.append(c);
					prevback = false;
				}else{
					prevback = true;
				}
			case CHAR_DOUBLEQUOTE:
				if(prevback) sb.append(c);
				prevback = false;
				break;
			default:
				prevback = false;
				sb.append(c);
			}
		}
		return sb.toString();
	}

	 
}
