package net.sf.corn.converter.json;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;

import net.sf.corn.converter.ParsingException;
import static net.sf.corn.converter.json.IJsonConstants.*;

/**
 * This class offers a convenient method to parse a json string to a json object
 * @author Serhat Dirik
 */
public class JsonStringParser {

	private static final char CHAR_DOUBLEQUOTE = '"';
	private static final char CHAR_WHITESPACE = ' ';
	private static final char CHAR_SINGLEQUOTE = '\'';

	private int index;
	private Reader reader;
	private char lastChar;
	private boolean useLastChar;

	/**
	 * Construct a JSONTokener from a string.
	 * 
	 * @param s
	 *            A source string.
	 */
	private JsonStringParser(String s) {
		Reader reader = new StringReader(s);
		this.reader = reader.markSupported() ? reader : new BufferedReader(
				reader);
		this.useLastChar = false;
		this.index = 0;
	}

	

	/**
	 * Parses a json string and builds a JsonObject from it
	 * 
	 * @param jsonString
	 *            A String that follows json
	 * @return JsTypeObject that converted form of the given string
	 * @throws ParsingException
	 *             Throwns if parsing not succeeded
	 */
	public static JsTypeObject parseJsonString(String jsonString)
			throws ParsingException {
		String injson = jsonString.trim();

		if (injson == null) {
			return JsTypeNull.NULL;
		} else if (injson.equals(JsTypeObject.NULL_STRING_VALUE)) {
			return JsTypeNull.NULL;
		} else {
			JsonStringParser tokenizer = new JsonStringParser(injson);
			JsTypeObject retVal;
			retVal = tokenizer.nextValue();
			return retVal;
		}
	}

	static JsTypeObject parseComplex(JsonStringParser tokenizer)
			throws ParsingException {
		JsTypeComplex complex = new JsTypeComplex();
		char c;
		String key = null;

		if (tokenizer.nextClean() != CHAR_CLASS_START) {
			throw tokenizer
					.syntaxError("A JsonComplex text must begin with '{'");
		}

		boolean finalized = false;
		while (!finalized) {

			c = tokenizer.nextClean();
			switch (c) {
			case 0:
				throw tokenizer
						.syntaxError("A JsonComplex text must end with '}'");
			case CHAR_CLASS_END:
				finalized = true;
				break;
			default:
				tokenizer.back();
			}
			if (!finalized) {
				JsTypeObject jsObj = tokenizer.nextValue();
				if (!jsObj.isSimple() || jsObj.isNull()) {
					throw tokenizer
							.syntaxError("A JsonComplex must contain a key");
				}
				key = jsObj.toStringValue();

				/*
				 * The key is followed by ':'. We will also tolerate '=' or
				 * '=>'.
				 */

				c = tokenizer.nextClean();
				if (c == '=') {
					if (tokenizer.next() != '>') {
						tokenizer.back();
					}
				} else if (c != ':') {
					throw tokenizer.syntaxError("Expected a ':' after a key");
				}
				JsTypeObject value = tokenizer.nextValue();
				if (key.equals(ATTR_CLASSNAME)) {
					complex.setJavaClassName(value.toStringValue());
				} else {
					complex.put(key, value);
				}

				/*
				 * Pairs are separated by ','. We will also tolerate ';'.
				 */

				switch (tokenizer.nextClean()) {
				case ';':
				case ',':
					if (!(tokenizer.nextClean() == '}')) {
						tokenizer.back();
					}
					break;
				case '}':
					finalized = true;
					break;
				default:
					throw tokenizer.syntaxError("Expected a ',' or '}'");
				}
			}
		}
		// check if it's simple or collection indeed

		if (complex.size() == 0) {
			return complex.getJavaClassName() == null ? JsTypeNull.NULL
					: new JsTypeSimple(null, complex.getJavaClassName());
		} else if (complex.size() == 1
				&& complex.keys().iterator().next().equals(ATTR_VALUE)) {
			JsTypeObject single = complex.get(ATTR_VALUE);
			if (single.isSimple() && single.getJavaClassName() == null) {
				single.setJavaClassName(complex.getJavaClassName());
				return single;
			}
		} else if (complex.size() == 1
				&& complex.keys().iterator().next().equals(ATTR_LIST)) {
			JsTypeObject single = complex.get(ATTR_LIST);
			if (single.isList() && single.getJavaClassName() == null) {
				single.setJavaClassName(complex.getJavaClassName());
				return single;
			}
		}
		return complex;

	}

	static JsTypeObject parseList(JsonStringParser tokenizer) throws ParsingException {
		JsTypeList jsList = new JsTypeList();
		char c = tokenizer.nextClean();
		char q;
		if (c == '[') {
			q = ']';
		} else if (c == '(') {
			q = ')';
		} else {
			throw tokenizer.syntaxError("A JSONArray text must start with '['");
		}
		if (tokenizer.nextClean() == ']') {
			return jsList;
		}
		tokenizer.back();
		for (;;) {
			if (tokenizer.nextClean() == ',') {
				tokenizer.back();
				jsList.add(JsTypeNull.NULL);
			} else {
				tokenizer.back();
				jsList.add(tokenizer.nextValue());
			}
			c = tokenizer.nextClean();
			switch (c) {
			case ';':
			case ',':
				if (tokenizer.nextClean() == ']') {
					return jsList;
				}
				tokenizer.back();
				break;
			case ']':
			case ')':
				if (q != c) {
					throw tokenizer.syntaxError("Expected a '"
							+ new Character(q) + "'");
				}
				return jsList;
			default:
				throw tokenizer.syntaxError("Expected a ',' or ']'");
			}
		}
	}

	
	/**
	 * Back up one character. This provides a sort of lookahead capability, so
	 * that you can test for a digit or letter before attempting to parse the
	 * next number or identifier.
	 */
	void back() throws ParsingException {
		if (useLastChar || index <= 0) {
			throw new ParsingException(
					"Stepping back two steps is not supported");
		}
		index -= 1;
		useLastChar = true;
	}

	/**
	 * Get the next character in the source string.
	 * 
	 * @return The next character, or 0 if past the end of the source string.
	 */
	char next() throws ParsingException {
		if (this.useLastChar) {
			this.useLastChar = false;
			if (this.lastChar != 0) {
				this.index += 1;
			}
			return this.lastChar;
		}
		int c;
		try {
			c = this.reader.read();
		} catch (IOException exc) {
			throw new ParsingException(exc);
		}

		if (c <= 0) { // End of stream
			this.lastChar = 0;
			return 0;
		}
		this.index += 1;
		this.lastChar = (char) c;
		return this.lastChar;
	}

	/**
	 * Get the next n characters.
	 * 
	 * @param n
	 *            The number of characters to take.
	 * @return A string of n characters.
	 * @throws ParsingException
	 *             Substring bounds error if there are not n characters
	 *             remaining in the source string.
	 */
	String next(int n) throws ParsingException {
		if (n == 0) {
			return "";
		}

		char[] buffer = new char[n];
		int pos = 0;

		if (this.useLastChar) {
			this.useLastChar = false;
			buffer[0] = this.lastChar;
			pos = 1;
		}

		try {
			int len;
			while ((pos < n)
					&& ((len = reader.read(buffer, pos, n - pos)) != -1)) {
				pos += len;
			}
		} catch (IOException exc) {
			throw new ParsingException(exc);
		}
		this.index += pos;

		if (pos < n) {
			throw syntaxError("Substring bounds error");
		}

		this.lastChar = buffer[n - 1];
		return new String(buffer);
	}

	/**
	 * Get the next char in the string, skipping whitespace.
	 * 
	 * @throws ParsingException
	 * @return A character, or 0 if there are no more characters.
	 */
	char nextClean() throws ParsingException {
		for (;;) {
			char c = next();
			if (c == 0 || c > ' ') {
				return c;
			}
		}
	}

	/**
	 * Return the characters up to the next close quote character. Backslash
	 * processing is done. The formal JSON format does not allow strings in
	 * single quotes, but an implementation is allowed to accept them.
	 * 
	 * @param quote
	 *            The quoting character, either <code>"</code>
	 *            &nbsp;<small>(double quote)</small> or <code>'</code>
	 *            &nbsp;<small>(single quote)</small>.
	 * @return A String.
	 * @throws ParsingException
	 *             Unterminated string.
	 */
	String nextString(char quote) throws ParsingException {
		char c;
		StringBuffer sb = new StringBuffer();
		for (;;) {
			c = next();
			switch (c) {
			case 0:
			case '\n':
			case '\r':
				throw syntaxError("Unterminated string");
			case '\\':
				c = next();
				switch (c) {
				case 'b':
					sb.append('\b');
					break;
				case 't':
					sb.append('\t');
					break;
				case 'n':
					sb.append('\n');
					break;
				case 'f':
					sb.append('\f');
					break;
				case 'r':
					sb.append('\r');
					break;
				case 'u':
					sb.append((char) Integer.parseInt(next(4), 16));
					break;
				case 'x':
					sb.append((char) Integer.parseInt(next(2), 16));
					break;
				default:
					sb.append(c);
				}
				break;
			default:

				if (c == quote) {
					return sb.toString();
				}
				sb.append(c);
			}
		}
	}

	/**
	 * Get the next value. The value can be a Boolean, Double, Integer,
	 * JSONArray, JSONObject, Long, or String, or the JSONObject.NULL object.
	 * 
	 * @throws ParsingException
	 *             If syntax error.
	 * 
	 * @return An object.
	 * @throws JsonOperationNotSupportedException
	 */
	JsTypeObject nextValue() throws ParsingException {
		char c = nextClean();
		String s;

		switch (c) {
		case CHAR_DOUBLEQUOTE:
		case CHAR_SINGLEQUOTE:
			// back();

			String st = nextString(c);
			if (st.equalsIgnoreCase(JsTypeObject.NULL_STRING_VALUE)) {
				return JsTypeNull.NULL;
			} else {
				JsTypeSimple ret = new JsTypeSimple();
				ret.setSimpleValue(JsTypeObject.quote(st));
				return ret;
			}
		case CHAR_CLASS_START:
			back();
			return parseComplex(this);
		case CHAR_ARRAY_START:
		case CHAR_OPEN_PARANTHESIS:
			back();
			return parseList(this);
		}

		/*
		 * Handle unquoted text. This could be the values true, false, or null,
		 * or it can be a number. An implementation (such as this one) is
		 * allowed to also accept non-standard forms.
		 * 
		 * Accumulate characters until we reach the end of the text or a
		 * formatting character.
		 */

		StringBuffer sb = new StringBuffer();
		while (c >= CHAR_WHITESPACE && ",:]}/\\\"[{;=#".indexOf(c) < 0) {
			sb.append(c);
			c = next();
		}
		back();

		s = sb.toString().trim();
		if (s.isEmpty()) {
			throw syntaxError("Missing value");
		}
		if (s.equalsIgnoreCase(JsTypeObject.NULL_STRING_VALUE)) {
			return JsTypeNull.NULL;
		} else {
			JsTypeSimple ret = new JsTypeSimple();
			ret.setSimpleValue(s);
			return ret;
		}
	}

	/**
	 * Make a JsonParseException to signal a syntax error.
	 * 
	 * @param message
	 *            The error message.
	 * @return A JsonParseException object, suitable for throwing
	 */
	ParsingException syntaxError(String message) {
		return new ParsingException(message + toString());
	}

	@Override
	public String toString() {
		return lastChar+" at " + index;
	}
}
